import importlib.util
import inspect
from typing import Dict
from typing import List

from pdtj.constants import GOOGLE_DOCSTRING_ELEMENTS


def process_args(func: object) -> dict:
    """
    Extracts the arguments and their types from a function's signature.

    Args:
        func: The function to extract the arguments from.

    Returns:
        A dictionary mapping each argument name to its type and default value (if any).
    """
    args_with_types = {}
    try:
        argspec = inspect.getfullargspec(func)
        defaults = argspec.defaults or []
        defaults = list(reversed(defaults))
        for index, args in enumerate(reversed(argspec.args)):
            if args != "self":
                if args in argspec.annotations.keys():
                    args_with_types[args + ":"] = {"type": argspec.annotations[args]}
                    if index < len(defaults):
                        args_with_types[args + ":"]["default"] = defaults[index]
                    else:
                        args_with_types[args + ":"]["default"] = "No default argument"

        return args_with_types

    except ValueError:
        return args_with_types

    except TypeError:
        return args_with_types


def update_args(pre_processed_dict: dict, args_pre_processed_dict: dict, arg_dict: dict) -> dict:
    """
    Update the `args_pre_processed_dict` with the values from `arg_dict` and return the updated `pre_processed_dict`.

    Args:
        pre_processed_dict:
            A dictionary containing the pre-processed information.
        args_pre_processed_dict:
            A dictionary containing the pre-processed information for arguments.
        arg_dict:
            A dictionary containing the argument names and their descriptions and types.

    Returns:
        A dictionary containing the updated pre-processed information.
    """

    # Remove the introduction key from the args_pre_processed_dict
    args_pre_processed_dict.pop("introduction")

    for arg in args_pre_processed_dict.keys():
        args_pre_processed_dict[arg] = {"description": args_pre_processed_dict[arg]}
        args_pre_processed_dict[arg].update(arg_dict[arg + ":"])

    pre_processed_dict["args"] = args_pre_processed_dict

    return pre_processed_dict


def get_elements_position(text: str, docstring_elements: List[str] = GOOGLE_DOCSTRING_ELEMENTS) -> Dict[int, str]:
    """
    Return a dictionary containing the position of each docstring element in the text.

    Args:
        text:
            The docstring to parse.
        docstring_elements:
            A list of docstring elements to search for.
            Defaults to GOOGLE_DOCSTRING_ELEMENTS.

    Returns:
        A dictionary containing the position of each docstring element in the text.
    """
    elements_position_index = []

    element_position = {0: "introduction"}
    for element in docstring_elements:
        position = text.find(element)

        if position != -1:
            element_position[position] = element
            elements_position_index.append(position)

    return {key: element_position[key] for key in sorted(element_position)}


def split_docstring_by_elements(text: str, elements_position: Dict[int, str]) -> Dict[str, str]:
    """
    Splits a given docstring into sections based on the positions of specified elements.

    Args:
        text:
            The docstring to be split.
        elements_position:
            A dictionary containing the positions of the elements to split the docstring by.

    Returns:
        A dictionary containing the sections of the split docstring, where the key is the name of the
         section and the value is the text of the section.
    """
    last_element = len(elements_position) - 1
    elements_list = list(elements_position.keys())
    result = {}
    if last_element != 0:
        for index, position in enumerate(elements_list):
            element = elements_position[position]
            if index == 0:
                result[element.replace(":", "").lower()] = text[: elements_list[index + 1]].replace(element, "")

            elif index == last_element:
                result[element.replace(":", "").lower()] = text[elements_list[index] :].replace(element, "")

            else:
                result[element.replace(":", "").lower()] = text[position : elements_list[index + 1]].replace(
                    element, ""
                )

    elif len(elements_position) == 1:
        for element in elements_position.values():
            result[element.replace(":", "").lower()] = text

    return result


def get_text_documentation_from_element(element: object) -> str:
    """
    This function returns the cleaned-up text documentation of an element.

    Args:
        element: The element to extract the documentation from.

    Returns:
        The cleaned-up text documentation of the element.
    """
    docstring = inspect.getdoc(element)
    if docstring:
        return inspect.cleandoc(docstring)
    else:
        return ""


def parse_object(object_to_parse: object) -> dict:
    """
    Parses the docstring of an object and returns a dictionary with the parsed elements.

    Args:
        object_to_parse:
            The object to parse.

    Returns:
        A dictionary with the parsed elements of the object's docstring.
    """
    raw_text_clean = get_text_documentation_from_element(object_to_parse)
    args_dict = process_args(object_to_parse)

    elements_position = get_elements_position(raw_text_clean)
    result = split_docstring_by_elements(raw_text_clean, elements_position)

    if "args" in result.keys():
        elements_position = get_elements_position(result["args"], list(args_dict.keys()))
        result_args = split_docstring_by_elements(result["args"], elements_position)
        result = update_args(result, result_args, args_dict)

    return result


def docstring_file_parser(file: str) -> dict:
    """
    Parses the docstrings of the classes and functions defined in the given Python file.

    Args:
        file:
            The name of the Python file to parse.

    Returns:
        A dictionary with the parsed information. The dictionary keys are the names
        of the classes or functions defined in the file, and the values are dictionaries
        with the parsed information for each class or function.

    Example:
        Given a file `example.py` containing a class `ExampleClass` and a function
        `example_function`:

        >>> docstring_file_parser('example.py')
        {
            'ExampleClass': {
                'introduction': 'This is the docstring of the ExampleClass.',
                'method1': {
                    'introduction': 'This is the docstring of method1.',
                    'args': {
                        'arg1': {
                            'description': 'The first argument.',
                            'type': 'str'
                        },
                        'arg2': {
                            'description': 'The second argument.',
                            'type': 'int'
                        }
                    },
                    'return': {
                        'description': 'The result of the method.',
                        'type': 'float'
                    }
                }
            },
            'example_function': {
                'introduction': 'This is the docstring of example_function.',
                'args': {
                    'arg1': {
                        'description': 'The first argument.',
                        'type': 'str'
                    },
                    'arg2': {
                        'description': 'The second argument.',
                        'type': 'int'
                    }
                },
                'return': {
                    'description': 'The result of the function.',
                    'type': 'float'
                }
            }
        }
    """
    try:
        mod = importlib.import_module(f"{file.replace('.py', '')}", package=None)
        members = inspect.getmembers(mod, predicate=lambda x: inspect.isclass(x) or inspect.isfunction(x))
        member_dict = dict(members)

        dict_result = {}
        for key in member_dict:
            dict_result[key] = parse_object(member_dict[key])

            if inspect.isclass(member_dict[key]):
                class_members = inspect.getmembers(member_dict[key], predicate=inspect.isfunction)
                class_members_dict = dict(class_members)
                for class_key in class_members_dict:
                    dict_result[key][class_key] = parse_object(class_members_dict[class_key])

        return dict_result

    except FileNotFoundError:
        return {}



import yaml
from yaml.loader import SafeLoader

try:
    # Open the file and load it
    with open("./.pdtj.yaml") as file:
        parameters = yaml.load(file, Loader=SafeLoader)

except FileNotFoundError:
    parameters = None



GOOGLE_DOCSTRING_ELEMENTS = [
    "Args:",
    "Attributes:",
    "Examples:",
    "Raises:",
    "Returns:",
    "Yields:",
]


REMOVE_ITEMS = [
    "__builtins__",
    "__cached__",
    "__doc__",
    "__file__",
    "__loader__",
    "__name__",
    "__package__",
    "__spec__",
    "__init__",
    "__pycache__",
    "__init__.py",
    "__version__.py",
]



import os


def check_dir_existence(path: str, file_name: str) -> bool:
    """
    Checks whether a file or directory exists at the specified path.

    Args:
        path:
            The path to the directory containing the file or directory to check.
        file_name:
            The name of the file or directory to check for existence.

    Returns:
        True if the file or directory exists, False otherwise.
    """
    return os.path.exists(os.path.join(path, file_name))


def create_single_dir(path: str, folder_name: str) -> None:
    """
    Creates a single directory with the given `folder_name` at the specified `path`.

    Args:
        path:
            The path to create the directory at.
        folder_name:
            The name of the directory to create.

    Returns:
        None
    """
    os.makedirs(os.path.join(path, folder_name))


def create_docs_and_pdtj_dir() -> None:
    """Create 'docs' and 'pdtj' directories if they do not already exist.

    This function checks whether the 'docs' and 'pdtj' directories exist in the current working directory.
     If they do not exist, the function creates them using the 'os.makedirs()' method.

    Args:
        None

    Returns:
        None
    """
    current_path = os.getcwd()
    for folder_name in ["docs", "pdtj"]:
        if not check_dir_existence(current_path, folder_name):
            create_single_dir(current_path, folder_name)

        current_path = os.path.join(current_path, folder_name)



import argparse
import json
import os

from pdtj import __version__
from pdtj.files_finder import files_parser
from pdtj.files_finder import get_directory_type
from pdtj.parameters import parameters
from pdtj.util import create_docs_and_pdtj_dir


def get_arguments():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--project_name", type=str, action="store", default=parameters["name"]
    )

    return vars(parser.parse_args())


def main():
    project_name = get_arguments()["project_name"]

    create_docs_and_pdtj_dir()

    directory_dict = {project_name: {"type": get_directory_type(f"../{project_name}")}}
    path = f"./{project_name}"
    upper_file = project_name
    file_parser_result = files_parser(path, upper_file, directory_dict)

    file_parser_result[project_name].update(
        {"shelve": parameters["shelve"], "sub-project": parameters["sub-project"]}
    )

    path = os.path.join(os.getcwd(), "docs")
    path = os.path.join(path, "pdtj")
    with open(os.path.join(path, f"{__version__.version}.json"), "w") as file:
        file.write(json.dumps(str(file_parser_result)))



import os
from typing import List

from pdtj.constants import REMOVE_ITEMS
from pdtj.docstring_handler import docstring_file_parser


def get_directory_type(possible_directory: str) -> str:
    """
    Determines the type of a given path.

    Args:
        possible_directory: A path to check for its type.

    Returns:
        The type of the given path. Either "directory" or "file".

    Example:
        >>> get_directory_type('/path/to/directory')
        'directory'
        >>> get_directory_type('/path/to/file.txt')
        'file'
    """
    return "directory" if os.path.isdir(possible_directory) else "file"


def list_directory(path: str) -> List[str]:
    """
    Returns a list of files and directories located in the specified path,
    excluding items in the REMOVE_ITEMS list.

    Args:
        path: The path to the directory to list.

    Returns:
        A list of files and directories in the directory, excluding REMOVE_ITEMS.
    """
    return list(set(os.listdir(path)).difference(REMOVE_ITEMS))


def loop_over_current_dictionary(module_path: str, result: dict) -> dict:
    """
    Given a module path and a dictionary to add, create nested dictionaries according to the module
    path and merge the result dictionary. Return the resulting nested dictionary.

    Args:
        module_path:
            A string representing the path of a module, separated by dots.
        result:
            A dictionary to merge with the nested dictionaries.

    Returns:
        A nested dictionary with the keys specified by the module path and the values specified
         by the result dictionary.
    """
    module_list = module_path.split(".")
    open_dictionary = {}
    for index, module in enumerate(reversed(module_list)):
        if index == 0:
            open_dictionary[module] = {}

        else:
            nested_dict = {module: open_dictionary}
            open_dictionary = nested_dict

    open_dictionary.update(result)
    return open_dictionary


def files_parser(path: str, upper_file: str, current_dictionary: dict) -> dict:
    """
    Recursively parses a directory and its subdirectories for Python files and extracts
    their docstrings as well as other metadata.

    Args:
        path:
            The path to the directory to be parsed.
        upper_file:
            The parent file name of the current directory.
        current_dictionary:
            A dictionary to store the metadata extracted from the files.

    Returns:
        The updated `current_dictionary` with the metadata extracted from the files.

    """
    update_dictionary = {upper_file: {"type": get_directory_type(path)}}

    files_in_directory = list_directory(path)
    for file_name in files_in_directory:
        current_path = f"{os.path.join(path, file_name)}"
        module_path = current_path.replace("./", "").replace("\\", ".").replace("/", ".")

        file_type = get_directory_type(current_path)
        if file_type == "directory":
            result = files_parser(current_path, file_name, current_dictionary)
            current_dictionary = loop_over_current_dictionary(module_path, result)

        elif file_type == "file":
            if file_name.endswith(".py"):
                python_type = "python file"
                docstring_dict = docstring_file_parser(module_path)

                update_dictionary[upper_file][file_name] = {"type": python_type}
                update_dictionary[upper_file][file_name].update(docstring_dict)

    current_dictionary.update(update_dictionary)
    return current_dictionary



import json
import os
import pathlib

import setuptools

from pdtj import __version__

long_description = (pathlib.Path(__file__).parent / "README.md").read_text()


def get_requirements_from_pipfile_lock(pipfile_lock=None):
    if pipfile_lock is None:
        pipfile_lock = os.path.join(os.path.dirname(os.path.realpath(__file__)), "Pipfile.lock")
    lock_data = json.load(open(pipfile_lock))
    retval = []
    for package_name, package_data in lock_data.get("default", {}).items():
        if package_data.get("file"):
            package_ref = f'{package_name} @ {package_data["file"]}'
        else:
            package_ref = f'{package_name}{package_data["version"]}'
        retval.append(package_ref)
    return retval


def generate_requirements_txt(pipfile_lock_requirements):
    for requirement in pipfile_lock_requirements:
        with open(os.path.join("requirements.txt"), "w") as file:
            file.writelines(requirement)


requirements = get_requirements_from_pipfile_lock()
generate_requirements_txt(requirements)

setuptools.setup(
    name="pdtj",
    version="0.3.4",
    license="MIT",
    author="Joao Paulo Euko",
    url="https://github.com/Joaopeuko/pdtj",
    keywords=["python", "docstring", "json"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "pdtj = pdtj.json_generator:main",
        ]
    },
)
